@@ -1,10 +1,9 @@
 from coba.backports.metadata import version, entry_points, PackageNotFoundError
-from coba.backports.typing import Literal, Protocol
+from coba.backports.typing import Literal
 
 __all__ = [
     'version',
     'entry_points',
     'PackageNotFoundError',
     'Literal',
-    'Protocol'
 ]
\ No newline at end of file
