@@ -2,16 +2,15 @@
 
 import json
 import time
-import collections.abc
 
 from numbers import Number
+from abc import ABC, abstractmethod
 from collections import Counter, OrderedDict, defaultdict
 from itertools import count, accumulate, chain
-from abc import ABC, abstractmethod
-
-from typing import Iterator, Sequence, Generic, TypeVar, Any, Tuple, Union, Dict
+from typing import Iterator, Sequence, Generic, TypeVar, Any, Tuple, Union, Mapping
 
 from coba.exceptions import CobaException
+from coba.primitives import Sparse, Dense, Categorical
 
 _T_out = TypeVar('_T_out', bound=Any, covariant=True)
 
@@ -220,6 +219,42 @@ class OneHotEncoder(Encoder[Tuple[int,...]]):
         except KeyError as e:
             raise CobaException(f"We were unable to find {e} in {list(self._onehots.keys())}")
 
+class CategoricalEncoder(Encoder[Categorical]):
+    """An Encoder implementation that turns incoming values into a one hot representation."""
+
+    def __init__(self, values: Sequence[str] = []) -> None:
+        """Instantiate a OneHotEncoder.
+
+        Args:
+            values: Provide the universe of values for encoding and set `is_fit==True`.
+        """
+
+        self._categoricals = {v: Categorical(v,values) for v in sorted(set(values)) } if values else None
+
+    @property
+    def is_fit(self) -> bool:
+        return self._categoricals is not None
+
+    def fit(self, values: Sequence[str]) -> 'CategoricalEncoder':
+        return CategoricalEncoder(values = values)
+
+    def encode(self, value: Any) -> Categorical:
+        try:
+            return self._categoricals[value]
+        except KeyError as e:
+            raise CobaException(f"We were unable to find {e} in the categorical values of {list(self._categoricals.keys())}.")
+        except TypeError:
+            raise CobaException("This encoder must be fit before it can be used.")
+
+    def encodes(self, values: Sequence[Any]) -> Sequence[Tuple[int,...]]:
+
+        try:
+            return list(map(self._categoricals.__getitem__,values))
+        except KeyError as e:
+            raise CobaException(f"We were unable to find {e} in the categorical values of {list(self._categoricals.keys())}.")
+        except (TypeError, AttributeError):
+            raise CobaException("This encoder must be fit before it can be used.")
+
 class FactorEncoder(Encoder[int]):
     """An Encoder implementation that turns incoming values into factor representation."""
 
@@ -285,30 +320,30 @@ class InteractionsEncoder:
         self.n           = 0
         self._constant   = sum(num_interactions)
         self._cross_pows = OrderedDict(zip(interactions,map(OrderedDict,map(Counter,str_interactions))))
-        self._ns_max_pow = { n:max(p.get(n,0) for p in self._cross_pows.values()) for n in set(''.join(str_interactions)) }
+        self._ns_max_pow = { n:int(max(p.get(n,0) for p in self._cross_pows.values())) for n in set(''.join(str_interactions)) }
 
-    def encode(self, **ns_raw_values: Union[str, float, Sequence[Union[str,float]], Dict[Union[str,int],Union[str,float]]]) -> Union[Sequence[float], Dict[str,float]]:
+    def encode(self, **ns_raw_values: Union[str, float, Sequence[Union[str,float]], Mapping[Union[str,int],Union[str,float]]]) -> Union[Sequence[float], Mapping[str,float]]:
 
         self.n+= 1
 
         ns_raw_values = { k:v if v is not None else [] for k,v in ns_raw_values.items() }
 
         is_str = lambda v: isinstance(v,str)
-        is_seq = lambda v: isinstance(v,collections.abc.Sequence) and not is_str(v)
-        is_map = lambda v: isinstance(v,collections.abc.Mapping)
+        is_seq = lambda v: isinstance(v,Dense)
+        is_map = lambda v: isinstance(v,Sparse)
 
         is_sparse_type = lambda f: is_map(f) or is_str(f)
         is_sparse_sequ = lambda f: is_seq(f) and any(map(is_sparse_type,f))
 
         is_sparse = any(is_sparse_type(v) or is_sparse_sequ(v) for v in ns_raw_values.values())
 
-        def make_dict(v) -> Dict[str,Union[str,float]]:
+        def make_dict(v) -> Mapping[str,Union[str,float]]:
             return v if is_map(v) else dict(zip(map(str,count()),v)) if is_seq(v) else { "0":v }
 
         def make_list(v) -> Sequence[Union[str,float]]:
             return v if is_seq(v) else [v]
 
-        def handle_str(v: Dict[str,Union[str,float]]) -> Dict[str,float]:
+        def handle_str(v: Mapping[str,Union[str,float]]) -> Mapping[str,float]:
             return { (f"{x}{y}" if is_str(y) else x):(1 if is_str(y) else y) for x,y in v.items() }
 
         start = time.time()
@@ -319,15 +354,18 @@ class InteractionsEncoder:
             ns_values = { ns:make_list(v) for ns,v in ns_raw_values.items() if ns in self._ns_max_pow}
         self.times[0] += time.time()-start
 
+        pows = self._pows
+        cross = self._cross
+
         if is_sparse:
             start = time.time()
-            key_pows = { ns: self._pows(list(ns_values[ns].keys()  ), max_pow) for ns, max_pow in self._ns_max_pow.items() }
-            val_pows = { ns: self._pows(list(ns_values[ns].values()), max_pow) for ns, max_pow in self._ns_max_pow.items() }
+            key_pows = { ns: pows(list(ns_values[ns].keys()  ), max_pow) for ns, max_pow in self._ns_max_pow.items() }
+            val_pows = { ns: pows(list(ns_values[ns].values()), max_pow) for ns, max_pow in self._ns_max_pow.items() }
             self.times[1] += time.time()-start
 
             start = time.time()
-            key_crosses = [ self._cross(key_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
-            val_crosses = [ self._cross(val_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
+            key_crosses = [ cross(key_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
+            val_crosses = [ cross(val_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
             self.times[2] += time.time()-start
 
             start = time.time()
@@ -339,11 +377,11 @@ class InteractionsEncoder:
             return encoded
         else:
             start = time.time()
-            val_pows = { ns: self._pows(ns_values[ns], max_pow) for ns, max_pow in self._ns_max_pow.items() }
+            val_pows = { ns: pows(ns_values[ns], max_pow) for ns, max_pow in self._ns_max_pow.items() }
             self.times[1] += time.time()-start
 
             start = time.time()
-            val_crosses = [ self._cross(val_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
+            val_crosses = [ cross(val_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
             self.times[2] += time.time()-start
 
             start = time.time()
@@ -354,11 +392,11 @@ class InteractionsEncoder:
 
             return encoded
 
-    def _pows(self, values, degree):
+    def _pows(self, values: Sequence[Union[str,float]], degree):
         #WARNING: This function has been extremely optimized. Please baseline performance before and after making any changes.
         #WARNING: You can find three existing performance tests in test_performance.
 
-        if values in [[],{}]: return []
+        if not values: return []
 
         starts = [1]*len(values)
         terms  = [['']] if isinstance(values[0],str) else [[1]]
@@ -381,7 +419,7 @@ class InteractionsEncoder:
 
         values = [ ns_pows[ns][p] for ns,p in cross_pow.items() ]
         cross  = values[0]
-
+        
         if isinstance(cross[0],str):
             for vs in values[1:]: cross = [ o+v for o in cross for v in vs ]
         else:
