@@ -1,28 +1,35 @@
 import collections.abc
 
+from collections import Counter
+from zipfile import ZipFile, BadZipFile
 from pathlib import Path
 from typing import Sequence, overload, Union, Iterable, Iterator, Any, Optional, Tuple, Callable, Mapping, Type
 from coba.backports import Literal
 
-from coba            import pipes
-from coba.contexts   import CobaContext, DiskCacher
-from coba.random     import CobaRandom
-from coba.pipes      import Pipes, Source, HttpSource, IterableSource, JsonDecode
-from coba.exceptions import CobaException
+from coba                 import pipes
+from coba.contexts        import CobaContext, DiskCacher
+from coba.primitives      import Context, Action, HashableSparse
+from coba.random          import CobaRandom
+from coba.pipes           import Pipes, Source, HttpSource, IterableSource, JsonDecode
+from coba.exceptions      import CobaException
+from coba.multiprocessing import CobaMultiprocessor
+from coba.learners        import Learner
 
-from coba.environments.filters   import EnvironmentFilter, Repr, Batch, Chunk
-from coba.environments.filters   import Binary, Shuffle, Take, Sparse, Reservoir, Cycle, Scale
+from coba.environments.primitives import Environment
+
+from coba.environments.templates  import EnvironmentsTemplateV1, EnvironmentsTemplateV2
+from coba.environments.openml     import OpenmlSimulation
+from coba.environments.synthetics import LinearSyntheticSimulation, NeighborsSyntheticSimulation
+from coba.environments.synthetics import KernelSyntheticSimulation, MLPSyntheticSimulation, LambdaSimulation
+from coba.environments.supervised import SupervisedSimulation
+
+from coba.environments.filters   import EnvironmentFilter, Repr, Batch, Chunk, Logged, Finalize, BatchSafe
+from coba.environments.filters   import Binary, Shuffle, Take, Sparse, Reservoir, Cycle, Scale, Unbatch
 from coba.environments.filters   import Impute, Where, Noise, Riffle, Sort, Flatten, Cache, Params, Grounded
-from coba.environments.templates import EnvironmentsTemplateV1, EnvironmentsTemplateV2
 
-from coba.environments.primitives import Environment, Context, Action
+from coba.environments.serialized import EnvironmentFromObjects, EnvironmentsToObjects, ZipMemberToObjects, ObjectsToZipMember
 
-from coba.environments.simulated.synthetics import LinearSyntheticSimulation, NeighborsSyntheticSimulation
-from coba.environments.simulated.synthetics import KernelSyntheticSimulation, MLPSyntheticSimulation, LambdaSimulation
-from coba.environments.simulated.openml     import OpenmlSimulation
-from coba.environments.simulated.supervised import SupervisedSimulation
-
-class Environments (collections.abc.Sequence):
+class Environments(collections.abc.Sequence, Sequence[Environment]):
     """A friendly wrapper around commonly used environment functionality."""
 
     @staticmethod
@@ -152,14 +159,17 @@ class Environments (collections.abc.Sequence):
     @staticmethod
     def from_openml(data_id: Union[int,Sequence[int]],
         drop_missing: bool = True,
-        take: int = None) -> 'Environments':
+        take: int = None,
+        *,
+        target:str = None) -> 'Environments':
         ...
 
     @overload
     @staticmethod
     def from_openml(*,task_id: Union[int,Sequence[int]],
         drop_missing: bool = True,
-        take: int = None) -> 'Environments':
+        take: int = None,
+        target:str = None) -> 'Environments':
         ...
 
     @staticmethod
@@ -203,6 +213,11 @@ class Environments (collections.abc.Sequence):
         """Create a SimulatedEnvironment from a supervised dataset"""
         return Environments(SupervisedSimulation(*args, **kwargs))
 
+    @staticmethod
+    def from_save(path:str) -> 'Environments':
+        make = lambda m: EnvironmentFromObjects(ZipMemberToObjects(path,m))
+        return Environments(list(map(make,ZipFile(path).namelist())))
+
     @overload
     def from_lambda(self,
         n_interactions: Optional[int],
@@ -268,9 +283,17 @@ class Environments (collections.abc.Sequence):
             seeds = range(kwargs['n'])
         else:
             seeds = flat(args) or [1]
-                
+
         if isinstance(seeds,int): seeds = [seeds]
-        return self.filter([Shuffle(seed) for seed in seeds])
+
+        shuffled = self.filter([Shuffle(seed) for seed in seeds])
+
+        #Experience has shown that most of the time we want to sort.
+        #This doesn't change the experiment results. It simply makes it
+        #easier monitor an experiment while it runs in the background.
+        ordered  = sorted(shuffled, key=lambda env: env.params['shuffle'])
+
+        return Environments(ordered)
 
     def sort(self, *keys: Union[str,int,Sequence[Union[str,int]]]) -> 'Environments':
         """Sort Environment interactions according to the context values indicated by keys."""
@@ -300,16 +323,22 @@ class Environments (collections.abc.Sequence):
     def scale(self,
         shift: Union[float,Literal["min","mean","med"]] = "min",
         scale: Union[float,Literal["minmax","std","iqr","maxabs"]] = "minmax",
-        target: Literal["context","rewards","argmax"] = "context",
+        targets: Union[Literal["context","rewards","argmax"], Sequence[Literal["context","rewards","argmax"]]] = "context",
         using: Optional[int] = None) -> 'Environments':
         """Apply an affine shift and scaling factor to precondition environments."""
-        return self.filter(Scale(shift, scale, target, using))
+        if isinstance(targets,str): targets = [targets]
+        return self.filter(Pipes.join(*[Scale(shift, scale, t, using) for t in targets]))
 
     def impute(self,
-        stat : Literal["mean","median","mode"] = "mean",
+        stats: Union[Literal["mean","median","mode"],Sequence[Literal["mean","median","mode"]]] = "mean",
+        indicator:bool = True,
         using: Optional[int] = None) -> 'Environments':
         """Impute missing values with a feature statistic using a given number of interactions."""
-        return self.filter(Impute(stat, using))
+        if isinstance(stats,str): stats = [stats]
+        envs = self
+        for stat in stats:
+            envs = self.filter(Impute(stat, indicator, using))
+        return envs
 
     def where(self,*,n_interactions: Union[int,Tuple[Optional[int],Optional[int]]] = None) -> 'Environments':
         """Only include environments which satisify the given requirements."""
@@ -329,7 +358,7 @@ class Environments (collections.abc.Sequence):
 
     def materialize(self) -> 'Environments':
         """Materialize the environments in memory. Ideal for stateful environments such as Jupyter Notebook."""
-        #we use pies.cache directly because the environment cache will copy 
+        #we use pipes.cache directly because the environment cache will copy 
         #which we don't need when we materialize an environment in memory
         envs = Environments([Pipes.join(env, pipes.Cache(25)) for env in self])
         for env in envs: list(env.read()) #force read to pre-load cache
@@ -354,6 +383,50 @@ class Environments (collections.abc.Sequence):
         envs = Environments([Pipes.join(env, Chunk()) for env in self])
         return envs.cache() if cache else envs
 
+    def logged(self, learners: Union[Learner,Sequence[Learner]]) -> 'Environments':
+        """Create a logged environment using the given learner for the logging policy."""
+        if not isinstance(learners, collections.abc.Sequence): learners = [learners]
+        return self.filter(BatchSafe(Finalize())).filter(list(map(Logged,learners)))
+
+    def unbatch(self):
+        """Unbatch interactions in the environments."""
+        return self.filter(Unbatch())
+
+    def save(self, path: str, processes:int=1, overwrite:bool=False) -> 'Environments':
+
+        if Path(path).exists():
+            try:
+                path_envs   = Environments.from_save(path)
+                path_params = [e.params for e in path_envs]
+                self_params = [e.params for e in self     ]
+                
+                try:
+                    for param in path_params:
+                        self_params.pop(self_params.index(param))
+                except ValueError:
+                    is_equal = False
+                else:
+                    is_equal = True
+
+                if is_equal:
+                    return path_envs
+                elif overwrite:
+                    Path(path).unlink()
+                else:
+                    raise CobaException("The Environments save file does not match the actual Environments and overwite is False.")
+            except BadZipFile:
+                if overwrite:
+                    Path(path).unlink()
+                else:
+                    raise CobaException("The given save file appears to be corruptted. Please check it and delete if it is unusable.")
+
+        if processes == 1:
+            Pipes.join(EnvironmentsToObjects(),ObjectsToZipMember(path)).write(self)
+        else:
+            Pipes.join(CobaMultiprocessor(EnvironmentsToObjects(),processes),ObjectsToZipMember(path)).write(self)
+
+        return Environments.from_save(path)
+
     def cache(self) -> 'Environments':
         """Create a cache point in the environments so that earlier steps in the pipeline can be re-used in several pipes."""
         return Environments([Pipes.join(env, Cache(25)) for env in self])
