@@ -1,15 +1,16 @@
 import time
 import pickle
 import warnings
+import copy
 
 from math import isnan
-from statistics import mean, median, stdev, mode
+from statistics import median, stdev, mode
 from numbers import Number
-from operator import eq, itemgetter, getitem, methodcaller
+from operator import eq, getitem, methodcaller, itemgetter
 from collections import defaultdict, deque
 from functools import lru_cache
 from itertools import islice, chain, tee, compress, repeat
-from typing import Hashable, Optional, Sequence, Union, Iterable, Dict, Any, List, Tuple, Callable, Mapping
+from typing import Optional, Sequence, Union, Iterable, Any, Tuple, Callable, Mapping
 from coba.backports import Literal
 
 from coba            import pipes, primitives
@@ -17,10 +18,11 @@ from coba.random     import CobaRandom
 from coba.exceptions import CobaException
 from coba.statistics import iqr
 from coba.utilities  import peek_first
-from coba.primitives import HashableSparse, HashableDense
+from coba.primitives import ScaleReward, BinaryReward, SequenceReward, BatchReward
+from coba.primitives import Feedback, BatchFeedback
+from coba.learners   import Learner, SafeLearner
 
 from coba.environments.primitives import EnvironmentFilter, Interaction
-from coba.environments.primitives import ScaleReward, BinaryReward, Feedback, SequenceReward
 
 class Identity(pipes.Identity, EnvironmentFilter):
     """Return whatever interactions are given to the filter."""
@@ -82,57 +84,41 @@ class Scale(EnvironmentFilter):
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
-        first, interactions = peek_first(interactions)
+        first, interactions = peek_first(Mutable().filter(interactions))
 
         if not interactions: return []
 
         remaining_interactions = iter(interactions)
-        fitting_interactions   = [ i.copy() for i in islice(remaining_interactions,self._using) ]
+        fitting_interactions   = list(islice(remaining_interactions,self._using))
 
         first_context = first['context']
         first_actions = first.get('actions')
 
-        if isinstance(first_context, primitives.Sparse) and self._shift != 0:
+        if isinstance(first_context, primitives.Sparse) and self._target=="context" and self._shift != 0:
             raise CobaException("Shift is required to be 0 for sparse environments. Otherwise the environment will become dense.")
 
         is_discrete       = first_actions and len(first_actions) > 0
         is_dense_context  = isinstance(first_context, primitives.Dense)
         is_sparse_context = isinstance(first_context, primitives.Sparse)
         is_value_context  = not (is_dense_context or is_sparse_context)
-        is_mutable        = isinstance(first['context'],(list,dict))
-
-        if self._target == "context" and is_dense_context:
-            possible_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float))]
-        elif self._target == "context" and is_sparse_context:
-            not_possible_cols = set([k for k,v in first['context'].items() if not isinstance(v,(int,float))])
-
-        start = time.time()
-        #we have to do this eventually, no reason to not do it upfront
-        if self._target == "context" and is_dense_context and not is_mutable:
-            for interaction in fitting_interactions:
-                interaction['context'] = list(interaction['context'])
-        elif self._target == "context" and is_sparse_context and not is_mutable:
-            for interaction in fitting_interactions:
-                #we are assuming that interaction['context'] is a LazySparse
-                #there's not really any other KV store that is not mutable
-                interaction['context'] = dict(interaction['context'].items())
-        self._times[0] += time.time()-start
 
         #get the values we wish to scale
         start = time.time()
         if self._target == "context" and is_dense_context:
-            if len(possible_cols) == 0:
+            potential_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float))]
+            if len(potential_cols) == 0:
                 unscaled = []
-            elif len(possible_cols) == 1:
-                unscaled = [list(map(itemgetter(*possible_cols),map(getitem,fitting_interactions,repeat("context"))))]
+            elif len(potential_cols) == 1:
+                unscaled = [ tuple(map(itemgetter(*potential_cols),map(getitem,fitting_interactions,repeat("context")))) ]
             else:
-                unscaled = list(zip(*map(itemgetter(*possible_cols),map(getitem,fitting_interactions,repeat("context")))))
+                unscaled = list(zip(*map(itemgetter(*potential_cols),map(getitem,fitting_interactions,repeat("context")))))
 
         elif self._target == "context" and is_sparse_context:
+            unscalable_cols = {k for k,v in first['context'].items() if not isinstance(v,(int,float))}
             unscaled = defaultdict(list)
             for interaction in fitting_interactions:
                 context = interaction['context']
-                for k in context.keys()-not_possible_cols:
+                for k in context.keys()-unscalable_cols:
                     unscaled[k].append(context[k])
 
         elif self._target == "context" and is_value_context:
@@ -142,6 +128,9 @@ class Scale(EnvironmentFilter):
             rwd_vals = lambda i: list(map(i['rewards'].eval,range(len(i['actions']))))
             unscaled = sum(list(map(rwd_vals,fitting_interactions)),[])
 
+        elif self._target == "rewards" and not is_discrete:
+            unscaled = []
+
         elif self._target == "argmax":
             unscaled = [interaction['rewards'].argmax() for interaction in fitting_interactions]
         self._times[1] += time.time()-start
@@ -149,11 +138,11 @@ class Scale(EnvironmentFilter):
         start = time.time()
         #determine the scale and shift values
         if self._target == "context" and is_dense_context:
-            shifts_scales = {}
-            for i,col in zip(possible_cols,unscaled):
+            shifts_scales = []
+            for i,col in zip(potential_cols,unscaled):
                 shift_scale = self._get_shift_and_scale(col)
                 if shift_scale is not None:
-                    shifts_scales[i] = shift_scale
+                    shifts_scales.append((i,)+shift_scale)
 
         elif self._target == "context" and is_sparse_context:
             shifts_scales = {}
@@ -173,44 +162,19 @@ class Scale(EnvironmentFilter):
 
         #now scale return
         elif self._target == "context" and is_dense_context:
-            for interaction in fitting_interactions:
-                new = interaction
-                if is_mutable:
-                    context = new['context'].copy()
-                    new['context'] = context
-                else:
-                    context = new['context']
-                for i, (shift,scale) in shifts_scales.items():
-                    context[i] = (context[i]+shift)*scale
-                yield new
-
-            for interaction in remaining_interactions:
-                new = interaction.copy()
-                if is_mutable:
-                    context = new['context'].copy()
-                    new['context'] = context
-                else: 
-                    context = list(new['context'])
-                    new['context'] = context
-                for i, (shift,scale) in shifts_scales.items():
-                    context[i] = (context[i]+shift)*scale
-                yield new
+            for interaction in chain(fitting_interactions, remaining_interactions):
+                context = interaction['context']
+                for i,shift,scale in shifts_scales:
+                     context[i] = (context[i]+shift)*scale
+                yield interaction
 
         elif self._target == "context" and is_sparse_context:
-            for interaction in fitting_interactions:
-                new = interaction
-                if is_mutable: new['context'] = new['context'].copy()
-                for k in shifts_scales.keys() & new['context'].keys():
+            for interaction in chain(fitting_interactions, remaining_interactions):
+                new = interaction # Mutable copies
+                context = new['context']
+                for k in shifts_scales.keys() & context.keys():
                     (shift,scale) = shifts_scales[k]
-                    new['context'][k] = (new['context'][k]+shift)*scale
-                yield new
-
-            for interaction in remaining_interactions:
-                new = interaction.copy()
-                new['context'] = new['context'].copy() if is_mutable else dict(new['context'])
-                for k in shifts_scales.keys() & new['context'].keys():
-                    (shift,scale) = shifts_scales[k]
-                    new['context'][k] = (new['context'][k]+shift)*scale
+                    context[k] = (context[k]+shift)*scale
                 yield new
 
         elif self._target == "context" and is_value_context:
@@ -257,7 +221,7 @@ class Scale(EnvironmentFilter):
         if shift == "min":
             return -min(values)
         elif shift == "mean":
-            return -mean(values)
+            return -sum(values)/len(values) #mean() is very slow due to calculations for precision
         elif shift == "med":
             return -median(values)
         return shift
@@ -287,75 +251,159 @@ class Impute(EnvironmentFilter):
 
     def __init__(self,
         stat : Literal["mean","median","mode"] = "mean",
+        indicator: bool = True,
         using: Optional[int] = None):
         """Instantiate an Impute filter.
 
         Args:
             stat: The statistic to use for impuatation.
+            indicator: Indicates whether a new binary feature should be added for missingness.
             using: The number of interactions to use to calculate the imputation statistics.
         """
 
         assert stat in ["mean","median","mode"]
 
         self._stat  = stat
+        self._miss  = indicator
         self._using = using
+        self._times = [0,0,0,0]
 
     @property
     def params(self) -> Mapping[str, Any]:
-        return { "impute_stat": self._stat, "impute_using": self._using }
+        return { "impute_stat": self._stat, "impute_using": self._using, "impute_indicator":self._miss }
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
-        first, interactions = peek_first(interactions)
+        first, interactions = peek_first(Mutable().filter(interactions))
+
+        if not interactions: return None
 
         is_dense  = isinstance(first['context'], primitives.Dense)
         is_sparse = isinstance(first['context'], primitives.Sparse)
+        is_value  = not is_dense and not is_sparse
 
-        iter_interactions  = iter(interactions)
-        train_interactions = list(islice(iter_interactions,self._using))
-        test_interactions  = chain.from_iterable([train_interactions, iter_interactions])
+        interactions       = iter(interactions)
+        using_interactions = list(islice(interactions,self._using))
+        other_interactions = interactions
 
-        stats   : Dict[Hashable,float]        = defaultdict(int)
-        features: Dict[Hashable,List[Number]] = defaultdict(list)
+        start = time.time()
+        if is_dense:
+            if self._stat in ["mean","median"]:
+                imputable_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float) or v is None)]
+            else:
+                imputable_cols = list(range(len(first['context'])))
 
-        for interaction in train_interactions:
-            for name,value in self._context_as_name_values(interaction['context']):
-                if isinstance(value,Number) and not isnan(value):
-                    features[name].append(value)
+        elif is_sparse:
+            if self._stat in ['mean','median']:
+                unimputable_cols = {k for k,v in first['context'].items() if not isinstance(v,(int,float) or v is None)}
+            else:
+                unimputable_cols = {}
 
-        for feat_name, feat_numeric_values in features.items():
+        elif is_value:
+            if self._stat in ["mean","median"]:
+                imputable_cols = isinstance(first['context'],(int,float)) or (first['context'] is None)
+            else:
+                imputable_cols = True
+        self._times[0] += time.time()-start
 
-            if self._stat == "mean":
-                stats[feat_name] = mean(feat_numeric_values)
+        start = time.time()
+        #get unimputed values
+        if is_dense:
+            if len(imputable_cols) == 0:
+                unimputed = []
+            elif len(imputable_cols) == 1:
+                unimputed = [ tuple(map(itemgetter(*imputable_cols),map(getitem,using_interactions,repeat("context")))) ]
+            else:
+                unimputed = list(zip(*map(itemgetter(*imputable_cols),map(getitem,using_interactions,repeat("context")))))
 
-            if self._stat == "median":
-                stats[feat_name] = median(feat_numeric_values)
+        elif is_sparse:
+            unimputed = defaultdict(list)
+            for interaction in using_interactions:
+                context = interaction['context']
+                for k in context.keys()-unimputable_cols:
+                    unimputed[k].append(context[k])
 
-            if self._stat == "mode":
-                stats[feat_name] = mode(feat_numeric_values)
+        elif is_value:
+            unimputed = [ interaction['context'] for interaction in using_interactions ]
+        self._times[1] += time.time()-start
 
-        for interaction in test_interactions:
+        start = time.time()
+        #calculate imputation statistics
+        if is_dense:
+            imputations = {}
+            impute_binary = {}
+            for i, col in zip(imputable_cols,unimputed):
+                imputation = self._get_imputation(col)
+                if imputation is not None:
+                    imputations[i] = imputation
+                    if self._miss and any([c is None for c in col]): 
+                        impute_binary[i] = len(impute_binary)
+
+        elif is_sparse:
+            imputations = {}
+            impute_binary = {}
+            binary_template = {}
+            for k,col in unimputed.items():
+                imputation = self._get_imputation(col + [0]*(len(using_interactions)-len(col)))
+                if imputation is not None:
+                    imputations[k] = imputation
+                    if self._miss and any([c is None for c in col]):
+                        impute_binary[k] = f"{k}_is_missing"
+                        binary_template[f"{k}_is_missing"] = 0
+
+        elif is_value:
+            imputations = self._get_imputation(unimputed)
+            impute_binary = self._miss and any([c is None for c in unimputed])
+        self._times[2] += time.time()-start
+
+        start = time.time()
+        for interaction in chain(using_interactions,other_interactions):
 
-            new     = interaction.copy()
             context = interaction['context']
-            imputed = {k: stats[k] if isinstance(v,float) and isnan(v) else v for k,v in self._context_as_name_values(context)}
 
-            if is_sparse:
-                new['context'] = imputed
-            elif is_dense:
-                new['context'] = tuple(imputed[k] for k,_ in self._context_as_name_values(context))
-            elif imputed:
-                new['context'] = imputed[0]
+            if is_dense:
+                is_missing = [0]*len(impute_binary)
+                for k,v in enumerate(context):
+                    if v is None and k in imputations: 
+                        context[k] = imputations[k]
+                        if k in impute_binary: 
+                            is_missing[impute_binary[k]] = 1
+                context += is_missing
 
-            yield new
+            elif is_sparse:
 
-    def _context_as_name_values(self,context) -> Sequence[Tuple[Hashable,Any]]:
+                is_missing = binary_template.copy()
+                for k,v in context.items():
+                    if v is None: 
+                        context[k] = imputations[k]
+                        if k in impute_binary:
+                            is_missing[impute_binary[k]] = 1
+                context.update(is_missing)
 
-        if isinstance(context,dict ): return context.items()
-        if isinstance(context,tuple): return enumerate(context)
-        if context is not None      : return [(0,context)]
+            elif is_value:
+                if impute_binary:
+                    if context is None: 
+                        interaction["context"] = [imputations,1]
+                    else:
+                        interaction["context"] = [context,0]
+                else:
+                    if context is None:
+                        interaction["context"] = imputations
 
-        return []
+            yield interaction
+        self._times[3] += time.time()-start
+
+    def _get_imputation(self,values):
+        try:
+            values = [v for v in values if v is not None]
+            if self._stat == "mean":
+                return sum(values)/len(values)
+            if self._stat == "median":
+                return median(values)
+            if self._stat == "mode":
+                return mode(values)
+        except:
+            return None
 
 class Sparse(EnvironmentFilter):
     """Sparsify an environment's feature representation.
@@ -566,53 +614,6 @@ class Where(EnvironmentFilter):
 
         return chain(taken_interactions, interactions)
 
-class Warm(EnvironmentFilter):
-    """Turn a SimulatedEnvironment into a WarmStartEnvironment."""
-
-    def __init__(self, n_warm:int, seed:int = 1):
-        """Instantiate a Warm filter.
-
-        Args:
-            n_warm: The number of interactions that should be turned into LoggedInteractions.
-            seed: The random number seed that determines the random logging policy for LoggedInteractions.
-        """
-        self._n_warm = n_warm
-        self._seed = seed
-
-    @property
-    def params(self) -> Mapping[str, Any]:
-        return { "n_warm": self._n_warm }
-
-    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
-
-        self._rng = CobaRandom(self._seed)
-
-        underlying_iterable    = iter(interactions)
-        logged_interactions    = map(self._to_logged_interaction, islice(underlying_iterable, self._n_warm))
-        simulated_interactions = underlying_iterable
-
-        return chain(logged_interactions, simulated_interactions)
-
-    def _to_logged_interaction(self, interaction: Interaction) -> Interaction:
-        new = interaction.copy()
-
-        actions       = interaction['actions']
-        rewards       = interaction['rewards']
-        num_actions   = len(actions)
-        probabilities = [1/num_actions] * num_actions
-
-        idx    = self._rng.choice(list(range(num_actions)), probabilities)
-        action = actions[idx]
-        prob   = probabilities[idx]
-        reward = rewards[idx]
-
-        new['type'] = 'logged'
-        new['action'] = action
-        new['probability'] = prob
-        new['reward'] = reward
-
-        return new
-
 class Riffle(EnvironmentFilter):
     """Riffle shuffle Interactions by taking actions from the end and evenly distributing into the beginning."""
 
@@ -813,7 +814,6 @@ class Grounded(EnvironmentFilter):
             argmax        = interaction['rewards'].argmax()
 
             new = interaction.copy()
-            new['type'] = 'grounded'
 
             if not is_binary_rwd:
                 new['rewards'] = BinaryReward(argmax)
@@ -886,8 +886,11 @@ class Batch(EnvironmentFilter):
         return {'batched': self._batch_size}
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
+
         for batch in self._batched(interactions, self._batch_size):
             new = { k: primitives.Batch(i[k] for i in batch) for k in batch[0] }
+            if 'rewards' in new: new['rewards'] = BatchReward(new['rewards'])
+            if 'feedbacks' in new: new['feedbacks'] = BatchFeedback(new['feedbacks'])
             yield new
 
     def _batched(self, iterable, n):
@@ -899,6 +902,25 @@ class Batch(EnvironmentFilter):
             yield batch
             batch = list(islice(it, n))
 
+class Unbatch(EnvironmentFilter):
+
+    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
+
+        first, interactions = peek_first(interactions)
+        if not interactions: return []
+        is_batched = isinstance(first[list(first.keys())[0]], primitives.Batch)
+
+        if not is_batched:
+            yield from interactions
+        else:
+            yield from self._unbatch(interactions)
+
+    def _unbatch(self, interactions: Iterable[Interaction]):
+        for interaction in interactions:
+            batch_size = len(interaction[list(interaction.keys())[0]])
+            for i in range(batch_size):
+                yield { k: interaction[k][i] for k in interaction }
+
 class BatchSafe(EnvironmentFilter):
 
     def __init__(self, filter: EnvironmentFilter) -> None:
@@ -906,25 +928,19 @@ class BatchSafe(EnvironmentFilter):
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
         first, interactions = peek_first(interactions)
-        
+
         if not interactions: return []
-        
+
         is_batched = isinstance(first[list(first.keys())[0]], primitives.Batch)
 
         if not is_batched:
-            return self._filter.filter(interactions)
+            yield from self._filter.filter(interactions)
         else:
             batch_size = len(first[list(first.keys())[0]])
-            debatched  = self._debatch(interactions)
-            filtered   = self._filter.filter(debatched)
-            rebatched  = Batch(batch_size).filter(filtered)
-            return rebatched
-
-    def _debatch(self, interactions: Iterable[Interaction]):
-        for interaction in interactions:
-            batch_size = len(interaction[list(interaction.keys())[0]])
-            for i in range(batch_size):
-                yield { k: interaction[k][i] for k in interaction }
+            unbatched = Unbatch().filter(interactions)
+            filtered  = self._filter.filter(unbatched)
+            rebatched = Batch(batch_size).filter(filtered)
+            yield from rebatched
 
 class Finalize(EnvironmentFilter):
 
@@ -938,10 +954,13 @@ class Finalize(EnvironmentFilter):
         first_has_actions = 'actions' in first and first['actions']
         first_has_action  = 'action'  in first
 
+        context_materialized = not first_has_context or isinstance(first['context']   ,(list,tuple,dict))
+        actions_materialized = not first_has_actions or isinstance(first['actions'][0],(list,tuple,dict))
+        action_materialized  = not first_has_action  or isinstance(first['action' ]   ,(list,tuple,dict))
+
         if first_has_context:
             is_dense_context  = isinstance(first['context'],primitives.Dense)
             is_sparse_context = isinstance(first['context'],primitives.Sparse)
-
         if first_has_actions:
             is_dense_action  = isinstance(first['actions'][0],primitives.Dense)
             is_sparse_action = isinstance(first['actions'][0],primitives.Sparse)
@@ -953,23 +972,20 @@ class Finalize(EnvironmentFilter):
 
             new = interaction.copy()
 
-            if first_has_context:
-                if is_dense_context:
-                    new['context'] = HashableDense(new['context'])
-                elif is_sparse_context:
-                    new['context'] = HashableSparse(new['context'])
+            if not context_materialized and is_dense_context:
+                new['context'] = list(new['context'])
+            elif not context_materialized and is_sparse_context :
+                    new['context'] = new['context'].copy()
 
-            if first_has_actions:
-                if is_dense_action:
-                    new['actions'] = list(map(HashableDense,new['actions']))
-                elif is_sparse_action:
-                    new['actions'] = list(map(HashableSparse,new['actions']))
+            if not actions_materialized and is_dense_action:
+                new['actions'] = list(map(list,new['actions']))
+            elif not actions_materialized and is_sparse_action:
+                new['actions'] = list(map(methodcaller('copy'),new['actions']))
 
-            if first_has_action:
-                if is_dense_action:
-                    new['action'] = HashableDense(new['action'])
-                elif is_sparse_action:
-                    new['action'] = HashableSparse(new['action'])
+            if not action_materialized and is_dense_action:
+                new['action'] = list(new['action'])
+            elif not action_materialized and is_sparse_action:
+                new['action'] = new['action'].copy()
 
             yield new
 
@@ -977,3 +993,79 @@ class Chunk(EnvironmentFilter):
     """A placeholder filter that exists only to semantically indicate how an environment pipe should be chunked for processing."""
     def filter(self, items: Iterable[Interaction]) -> Iterable[Interaction]:
         return items
+
+class Logged(EnvironmentFilter):
+
+    def __init__(self, learner: Learner) -> None:
+        self._learner = learner
+
+    @property
+    def params(self) -> Mapping[str, Any]:
+        return {"learner": self._learner.params}
+
+    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
+
+        learner = SafeLearner(copy.deepcopy(self._learner))
+
+        first, interactions = peek_first(interactions)
+
+        predict = learner.predict
+        learn   = learner.learn
+
+        if 'context' not in first or 'actions' not in first or 'rewards' not in first:
+            raise CobaException("We were unable to create a logged representation of the interaction.")
+
+        for interaction in interactions:
+
+            context = interaction['context']
+            actions = interaction['actions']
+            rewards = interaction['rewards']
+
+            action,prob,info = predict(context, actions)
+
+            reward = rewards.eval(action)
+
+            learn(context, actions, action, reward, prob, **info)
+
+            new = interaction.copy()
+
+            new['action']      = action
+            new['probability'] = prob
+            new['reward']      = reward
+
+            yield new
+
+class Mutable(EnvironmentFilter):
+
+    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
+        
+        first, interactions = peek_first(interactions)
+        
+        if not interactions: return []
+
+        first_is_dense   = isinstance(first['context'], primitives.Dense)
+        first_is_sparse  = isinstance(first['context'], primitives.Sparse)
+        first_is_value   = not (first_is_dense or first_is_sparse)
+        first_is_mutable = isinstance(first['context'], (list,dict))
+
+        if first_is_mutable:
+            for interaction in interactions:
+                new = interaction.copy()
+                new['context'] = new['context'].copy()
+                yield new
+                            
+        elif first_is_dense:
+            for interaction in interactions:
+                new = interaction.copy()
+                new['context'] = list(new['context'])
+                yield new
+        
+        elif first_is_sparse:
+            for interaction in interactions:
+                new = interaction.copy()
+                new['context'] = dict(new['context'].items())
+                yield new
+
+        elif first_is_value:
+            for interaction in interactions:
+                yield interaction.copy()
