@@ -1,9 +1,10 @@
 from pathlib import Path
-from typing import Sequence, Optional, Union
+from itertools import product
+from typing import Sequence, Optional, Union, overload, Tuple
 
-from coba.pipes import Pipes, Foreach
+from coba.pipes import Pipes
 from coba.learners import Learner
-from coba.environments import Environment, Environments
+from coba.environments import Environment
 from coba.multiprocessing import CobaMultiprocessor
 from coba.contexts import CobaContext, ExceptLog, StampLog, NameLog, DecoratedLogger, NullLogger
 from coba.exceptions import CobaException
@@ -16,10 +17,12 @@ from coba.experiments.results import Result, TransactionIO
 class Experiment:
     """An Experiment using a collection of environments and learners."""
 
+    @overload
     def __init__(self,
         environments    : Union[Environment, Sequence[Environment]],
         learners        : Union[Learner,Sequence[Learner]],
-        description     : str = None,
+        *,
+        description     : str             = None,
         learner_task    : LearnerTask     = SimpleLearnerInfo(),
         environment_task: EnvironmentTask = SimpleEnvironmentInfo(),
         evaluation_task : EvaluationTask  = SimpleEvaluation()) -> None:
@@ -34,25 +37,44 @@ class Experiment:
             evaluation_task: A task which evaluates a learner on an environment.
         """
 
-        if (not isinstance(environments, (Environments, list, tuple))):
-            environments = [environments]
+    @overload
+    def __init__(self,
+        eval_pairs      : Sequence[Tuple[Learner,Environment]],
+        *,
+        description     : str             = None,
+        learner_task    : LearnerTask     = SimpleLearnerInfo(),
+        environment_task: EnvironmentTask = SimpleEnvironmentInfo(),
+        evaluation_task : EvaluationTask  = SimpleEvaluation()) -> None:
+        ...
+        """Instantiate an Experiment.
 
-        if (not isinstance(learners, (list, tuple))):
-            learners = [learners]
+        Args:
+            eval_pairs: The collection of learners with their evaluation environments.
+            learner_task: A task which describes a learner.
+            environment_task: A task which describes an environment.
+            evaluation_task: A task which evaluates a learner on an environment.
+        """
 
-        self._environments     = environments
-        self._learners         = learners
-        self._description      = description
-        self._learner_task     = learner_task
-        self._environment_task = environment_task
-        self._evaluation_task  = evaluation_task
+    def __init__(self, *args, **kwargs) -> None:
+        """Instantiate an Experiment."""
 
-        if any([env is None for env in environments]):
-            raise CobaException("An Environment was given whose value was None, which can't be processed.")
+        if len(args) == 2:
+            envs = [args[0]] if hasattr(args[0],'read') else args[0]
+            lrns = [args[1]] if hasattr(args[1],'predict') else args[1]
+            args = [list(zip(*reversed(list(zip(*product(envs,lrns))))))]
 
-        if any([lrn is None for lrn in learners]):
+        self._pairs            = args[0]
+        self._description      = kwargs.get('description',None)
+        self._learner_task     = kwargs.get('learner_task', SimpleLearnerInfo())
+        self._environment_task = kwargs.get('environment_task', SimpleEnvironmentInfo())
+        self._evaluation_task  = kwargs.get('evaluation_task', SimpleEvaluation())
+
+        if any([lrn is None for lrn,_ in self._pairs]):
             raise CobaException("A Learner was given whose value was None, which can't be processed.")
 
+        if any([env is None for _,env in self._pairs]):
+            raise CobaException("An Environment was given whose value was None, which can't be processed.")
+
         self._processes        : Optional[int] = None
         self._maxchunksperchild: Optional[int] = None
         self._maxtasksperchunk : Optional[int] = None
@@ -120,17 +142,17 @@ class Experiment:
         if result_file and Path(result_file).exists():
             CobaContext.logger.log("Restoring existing experiment logs...")
             restored = Result.from_file(result_file)
-        else: 
+        else:
             restored = None
 
-        n_given_learners     = len(self._learners)
-        n_given_environments = len(self._environments)
+        n_given_learners     = len(set([l for l,_ in self._pairs]))
+        n_given_environments = len(set([e for _,e in self._pairs]))
 
         if restored:
             assert n_given_learners     == restored.experiment.get('n_learners',n_given_learners)        , "The current experiment doesn't match the given transaction log."
             assert n_given_environments == restored.experiment.get('n_environments',n_given_environments), "The current experiment doesn't match the given transaction log."
 
-        workitems  = CreateWorkItems(self._environments, self._learners, self._learner_task, self._environment_task, self._evaluation_task)
+        workitems  = CreateWorkItems(self._pairs, self._learner_task, self._environment_task, self._evaluation_task)
         unfinished = RemoveFinished(restored)
         chunk      = ChunkByChunk()
         max_chunk  = MaxChunkSize(mt)
@@ -141,9 +163,10 @@ class Experiment:
         process        = multi_process if mp > 1 or mc != 0 else single_process
 
         try:
-            if not restored: sink.write(["T0", {'n_learners':n_given_learners, 'n_environments':n_given_environments, 'description':self._description }])
-            Pipes.join(workitems, unfinished, process, Foreach(sink)).run()
-        except KeyboardInterrupt as e: # pragma: no cover
+            if not restored: sink.write([["T0", {'n_learners':n_given_learners, 'n_environments':n_given_environments, 'description':self._description }]])
+            Pipes.join(workitems, unfinished, process, sink).run()
+
+        except KeyboardInterrupt: # pragma: no cover
             CobaContext.logger.log("Experiment execution was manually aborted via Ctrl-C")
 
         except Exception as ex: # pragma: no cover
