@@ -427,8 +427,9 @@ class TransactionIO(Source['Result'], Sink[Any]):
         else:
             raise CobaException("We were unable to determine the appropriate Transaction reader for the file.")
 
-    def write(self, transaction: Any) -> None:
-        self._transactionIO.write(transaction)
+    def write(self, transactions: Iterable[Any]) -> None:
+        for transaction in transactions:
+            self._transactionIO.write(transaction)
 
     def read(self) -> 'Result':
         return self._transactionIO.read()
@@ -451,8 +452,8 @@ class Plotter:
         title: str,
         xlabel: str,
         ylabel: str,
-        xlim: Optional[Tuple[Number,Number]],
-        ylim: Optional[Tuple[Number,Number]],
+        xlim: Tuple[Optional[Number],Optional[Number]],
+        ylim: Tuple[Optional[Number],Optional[Number]],
         xticks: bool,
         yticks: bool,
         out: Union[None,Literal['screen'],str]) -> None:
@@ -466,8 +467,8 @@ class MatplotlibPlotter(Plotter):
         title: str,
         xlabel: str,
         ylabel: str,
-        xlim: Optional[Tuple[Number,Number]],
-        ylim: Optional[Tuple[Number,Number]],
+        xlim: Tuple[Optional[Number],Optional[Number]],
+        ylim: Tuple[Optional[Number],Optional[Number]],
         xticks: bool,
         yticks: bool,
         xrotation: Optional[float],
@@ -475,6 +476,9 @@ class MatplotlibPlotter(Plotter):
         out: Union[None,Literal['screen'],str]
     ) -> None:
 
+        xlim = xlim or [None,None]
+        ylim = ylim or [None,None]
+
         PackageChecker.matplotlib('Result.plot_learners')
         import matplotlib.pyplot as plt #type: ignore
 
@@ -519,7 +523,7 @@ class MatplotlibPlotter(Plotter):
                     XYE = [(x,y,e) for i,x,y,e in zip(count(),X,Y,YE) if in_lim(num_coalesce(x,i),xlim) and in_lim(num_coalesce(y,i),ylim)]
                     X,Y,YE = map(list,zip(*XYE)) if XYE else ([],[],[])
 
-                if isinstance(c,int): c = color_cycle[c]
+                if isinstance(c,int): c = color_cycle[c%len(color_cycle)]
 
                 not_err_bar = lambda E: not E or all([e is None for e in E])
 
@@ -538,18 +542,14 @@ class MatplotlibPlotter(Plotter):
             if yrotation is not None:
                 plt.yticks(rotation=yrotation)
 
-            padding = .05
-            ax.margins(0)
-            ax.set_xticks(ax.get_xticks() if not xlim else [min(ax.get_xlim()[1], max(ax.get_xlim()[0],x)) for x in ax.get_xticks()])
-            ax.margins(padding)
+            if xlim[0] is None or xlim[1] is None:
+                ax.autoscale(axis='x')
 
-            if xlim:
-                x_pad = padding*(xlim[1]-xlim[0])
-                ax.set_xlim(xlim[0]-x_pad, xlim[1]+x_pad)
+            if ylim[0] is None or ylim[1] is None:
+                ax.autoscale(axis='y')
 
-            if ylim:
-                y_pad = padding*(ylim[1]-ylim[0])
-                ax.set_ylim(ylim[0]-y_pad, ylim[1]+y_pad)
+            ax.set_xlim(*xlim)
+            ax.set_ylim(*ylim)
 
             ax.set_title(title, loc='left', pad=15)
             ax.set_ylabel(ylabel)
@@ -838,8 +838,8 @@ class Result:
         err        : Union[Literal['se','sd','bs'], None, PointAndInterval] = None,
         labels     : Sequence[str] = None,
         colors     : Sequence[str] = None,
-        xlim       : Tuple[Number,Number] = None,
-        ylim       : Tuple[Number,Number] = None,
+        xlim       : Tuple[Optional[Number],Optional[Number]] = None,
+        ylim       : Tuple[Optional[Number],Optional[Number]] = None,
         xticks     : bool = True,
         yticks     : bool = True,
         out        : Union[None,Literal['screen'],str] = 'screen',
@@ -868,6 +868,9 @@ class Result:
             ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
         """
 
+        xlim = xlim or [None,None]
+        ylim = ylim or [None,None]
+
         x = [x] if isinstance(x,str) else list(x)
         self._validate_parameters(x)
 
@@ -931,11 +934,12 @@ class Result:
         span  : int = None,
         err   : Union[Literal['se','sd','bs'], None, PointAndInterval] = None,
         labels: Sequence[str] = None,
-        colors: Union[int,Sequence[Union[str,int]]] = None,
-        xlim  : Tuple[Number,Number] = None,
-        ylim  : Tuple[Number,Number] = None,
+        colors: Sequence[Union[str,int]] = None,
+        xlim  : Tuple[Optional[Number],Optional[Number]] = None,
+        ylim  : Tuple[Optional[Number],Optional[Number]] = None,
         xticks: bool = True,
         yticks: bool = True,
+        top_n: int = None,
         out   : Union[None,Literal['screen'],str] = 'screen',
         ax = None) -> None:
         """Plot the performance of multiple learners on multiple environments. It gives a sense of the expected
@@ -957,10 +961,14 @@ class Result:
             ylim: Define the y-axis limits to plot. If `None` the y-axis limits will be inferred.
             xticks: Whether the x-axis labels should be drawn.
             yticks: Whether the y-axis labels should be drawn.
+            top_n: Only plot the top_n learners. If `None` all learners will be plotted.
             out: Indicate where the plot should be sent to after plotting is finished.
             ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
         """
 
+        xlim = xlim or [None,None]
+        ylim = ylim or [None,None]
+
         if isinstance(ids,int): ids = [ids]
         if isinstance(colors,int): colors = [colors]
         if isinstance(labels,str): labels = [labels]
@@ -979,22 +987,25 @@ class Result:
 
         style = "-" if x == ['index'] else "."
 
+        def get_color(colors:Sequence[Union[str,int]], i:int):
+            return i if not colors else i+max(colors) if isinstance(colors[0],int) else colors[i] if i < len(colors) else i
+
         for i, (lrn_id, lrn_rows) in enumerate(groupby(sorted(rows, key=get_key),key=get_key)):
             lrn_rows = list(lrn_rows)
             XYE      = TransformToXYE().filter(lrn_rows, env_rows, x, y, err)
-            color    = i if not colors else i+max(colors) if isinstance(colors[0],int) else colors[i] if i < len(colors) else i
-            label    = labels[i] if labels and i < len(labels) else self.learners[lrn_id]['full_name']            
+            color    = get_color(colors,i)
+            label    = labels[i] if labels and i < len(labels) else self.learners[lrn_id]['full_name']
             lines.append(Points(*zip(*XYE), color, 1, label, style))
 
-        lines  = sorted(lines, key=lambda line: -line[1][-1])
+        lines  = sorted(lines, key=lambda line: line[1][-1], reverse=True)
         xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
         ylabel = y.capitalize().replace("_pct"," Percent")
 
         title = ("Instantaneous" if span == 1 else f"Span {span}" if span else "Progressive") + f" {ylabel}"
         title = title + f" ({len(lrn_rows) if x==['index'] else len(XYE)} Environments)"
 
-        if x != ['index']:
-            title = "Final " + title
+        if x != ['index']: title = f"Final {title}"
+        if top_n         : lines = [l._replace(color=get_color(colors,i)) for i,l in enumerate(lines[:top_n]) ]
 
         self._plotter.plot(ax, lines, title, xlabel, ylabel, xlim, ylim, xticks, yticks, 0, 0, out)
 
