@@ -7,17 +7,17 @@ import warnings
 from abc import ABC, abstractmethod
 from statistics import mean
 from itertools import combinations
-from typing import Iterable, Any, Sequence, Mapping, Hashable, Union
+from typing import Iterable, Any, Sequence, Mapping, Hashable, Union, overload
+from coba.backports import Literal
 
 from coba.statistics import percentile
-from coba.exceptions import CobaExit, CobaException
+from coba.exceptions import CobaExit
 from coba.learners import Learner, SafeLearner
 from coba.encodings import InteractionsEncoder
 from coba.utilities import PackageChecker, peek_first
 from coba.contexts import CobaContext
 from coba.primitives import Batch
-from coba.environments import Environment, SafeEnvironment
-from coba.environments import Interaction, SimulatedInteraction, LoggedInteraction, GroundedInteraction
+from coba.environments import Environment, SafeEnvironment, Interaction
 
 class LearnerTask(ABC):
     """A task which describes a Learner."""
@@ -68,10 +68,12 @@ class EvaluationTask(ABC):
 
 class SimpleEvaluation(EvaluationTask):
 
+    @overload
     def __init__(self,
         reward_metrics: Union[str,Sequence[str]] = "reward",
         time_metrics: bool = False,
         probability: bool = False) -> None:
+        ...
         """
         Args:
             reward_metrics: The reward metrics to that should be recorded for each example.
@@ -80,129 +82,135 @@ class SimpleEvaluation(EvaluationTask):
             probability: Indicates whether the action's probability should be recorded.
         """
 
-        if not reward_metrics:
-            raise CobaException("At least one performance metric is required.")
+    @overload
+    def __init__(self, record: Sequence[Literal['reward','rank','regret','time','prob']] = ['reward']) -> None:
+        ...
+        """
+        Args:
+            record: The datapoints to record for each interaction.
+        """
 
-        if isinstance(reward_metrics,str): 
-            reward_metrics = [reward_metrics]
+    def __init__(self, *args, **kwargs) -> None:
 
-        self._metrics = reward_metrics
-        self._time    = time_metrics
-        self._prob    = probability
+        is_old_interface = (len(args) > 1 and isinstance(args[1],bool)) or (len(kwargs) > 0 and 'record' not in kwargs)
+
+        if not is_old_interface:
+            if not args and not kwargs: args = [['reward']]
+            record = args[0] if args else kwargs['record']
+            if isinstance(record,str): record = [record]
+        else:
+            record = []
+
+            if 'reward_metrics' in kwargs:
+                args = [kwargs['reward_metrics']]
+            if 'time_metrics' in kwargs:
+                args = ["reward"] if not args else list(args)
+                args.append(kwargs['time_metrics'])
+            if 'probability' in kwargs:
+                if len(args) == 1: args = list(args) + [False]
+                if len(args) == 0: args = ["reward", False]
+                args.append(kwargs['probability']) 
+
+            if len(args) > 0 and     isinstance(args[0],str): record.append(args[0])
+            if len(args) > 0 and not isinstance(args[0],str): record.extend(args[0])
+            if len(args) > 1 and args[1]                    : record.append("time")
+            if len(args) > 2 and args[2]                    : record.append("prob")
+
+        self._record = record
 
     def process(self, learner: Learner, interactions: Iterable[Interaction]) -> Iterable[Mapping[Any,Any]]:
 
         learner = SafeLearner(learner)
 
+        record_prob = 'prob' in self._record
+        record_time = 'time' in self._record
+
         first, interactions = peek_first(interactions)
 
         predict  = learner.predict
         learn    = learner.learn
 
         discrete = first and len(first.get('actions',[])) > 0
-        batched  = isinstance(first['type'],Batch)
 
         learning_info = CobaContext.learning_info
 
-        if 'rank' in self._metrics and not discrete:
+        if 'rank' in self._record and not discrete:
             warnings.warn(f"The rank metric can only be calculated for discrete environments")
 
-        calc_rank   = 'rank'   in self._metrics and discrete
-        calc_reward = 'reward' in self._metrics
-        calc_regret = 'regret' in self._metrics
+        calc_rank   = 'rank'   in self._record and discrete
+        calc_reward = 'reward' in self._record
+        calc_regret = 'regret' in self._record
 
         get_reward = lambda reward                    : reward
         get_regret = lambda reward, rewards           : rewards.max()-reward
         get_rank   = lambda reward, rewards, n_actions: sorted(map(rewards.eval,range(n_actions))).index(reward)/(n_actions-1)
 
+        learning_info.clear()
+
         for interaction in interactions:
 
-            learning_info.clear()
+            interaction = interaction.copy()
 
             out     = {}
-            
-            tipe    = interaction['type'][0] if batched else interaction['type']
+            context = interaction.pop('context')
+            batched = isinstance(context, Batch)
 
-            if tipe == 'simulated':
+            is_on_policy_eval   = 'actions' in interaction and 'rewards' in interaction
+            is_off_policy_learn = 'action' in interaction and 'reward' in interaction
+            is_on_policy_learn  = is_on_policy_eval and not is_off_policy_learn
 
-                context = interaction['context']
-                actions = interaction['actions']
-                rewards = interaction['rewards']
-                extra   = {k:interaction[k] for k in interaction.keys()-SimulatedInteraction.keywords }
+            is_predict = is_on_policy_eval
+            is_learn   = is_off_policy_learn or is_on_policy_learn
+
+            if is_predict:
+                actions = interaction.pop('actions')
+                rewards = interaction.pop('rewards')
 
                 start_time       = time.time()
                 action,prob,info = predict(context, actions)
                 predict_time     = time.time()-start_time
 
-                reward = list(map(lambda r,a: r.eval(a),rewards,action)) if batched else rewards.eval(action)
+                reward = rewards.eval(action)
+
+                if record_time: out['predict_time'] = predict_time
+                if record_prob: out['probability']  = prob
+
+                if not batched:
+                    if calc_reward : out['reward'] = get_reward(reward)
+                    if calc_regret : out['regret'] = get_regret(reward, rewards)
+                    if calc_rank   : out['rank'  ] = get_rank  (reward, rewards, len(actions))
+                elif batched:
+                    if calc_reward : out['reward'] = mean(map(get_reward,reward))
+                    if calc_regret : out['regret'] = mean(map(get_regret,reward,rewards))
+                    if calc_rank   : out['rank'  ] = mean(map(get_rank  ,reward,rewards,len(actions)))
+
+                if 'feedbacks' in interaction:
+                    reward = interaction.pop('feedbacks').eval(action)
+                    out['feedback'] = reward
+
+            if is_learn:
+
+                if is_off_policy_learn:
+                    actions = interaction.pop('actions',actions if is_predict else None)
+                    action  = interaction.pop('action')
+                    reward  = interaction.pop('reward')
+                    prob    = interaction.pop('probability',None)
+                    info    = {}
 
                 start_time = time.time()
                 learn(context, actions, action, reward, prob, **info)
                 learn_time = time.time() - start_time
 
-            elif tipe == 'logged':
+                if record_time: out['learn_time'] = learn_time
 
-                context = interaction['context']
-                actions = interaction.get('actions')
-                rewards = interaction.get('rewards')
-                extra   = {k:interaction[k] for k in interaction.keys()-LoggedInteraction.keywords }
+            if interaction:
+                out.update(interaction)
 
-                if not actions or not rewards:
-                    predict_time = None
-                    reward       = None
-                    prob         = None
-                    info         = None
-                else:
-                    start_time   = time.time()
-                    action,prob  = predict(context, actions)[:2]
-                    reward       = list(map(rewards.eval,action)) if batched else rewards.eval(action)
-                    predict_time = time.time()-start_time
-
-                start_time = time.time()
-                learn(context, actions, interaction['action'], interaction['reward'], interaction.get('probability'))
-                learn_time = time.time()-start_time
-
-            elif tipe == 'grounded':
-
-                context   = interaction['context']
-                actions   = interaction['actions']
-                rewards   = interaction['rewards']
-                feedbacks = interaction['feedbacks']
-                extra     = { k:interaction[k] for k in interaction.keys()-GroundedInteraction.keywords }
-
-                start_time       = time.time()
-                action,prob,info = predict(context, actions)
-                reward           = list(map(rewards.eval,action)) if batched else rewards.eval(action)
-                feedback         = list(map(feedbacks.eval,action)) if batched else feedbacks.eval(action)
-                predict_time     = time.time()-start_time
-
-                start_time = time.time()
-                learn(context, actions, action, feedback, prob, **info)
-                learn_time = time.time()-start_time
-
-                out['feedback'] = feedback
-
-            else:
-                raise CobaException("An unknown interaction type was received.")
-
-            if reward is not None:
-                if not batched:
-                    if calc_reward : out['reward'] = get_reward(reward)
-                    if calc_regret : out['regret'] = get_regret(reward, rewards)
-                    if calc_rank   : out['rank'  ] = get_rank  (reward, rewards, len(actions))
-                else:
-                    if calc_reward : out['reward'] = mean(map(get_reward,reward))
-                    if calc_regret : out['regret'] = mean(map(get_regret,reward,rewards))
-                    if calc_rank   : out['rank'  ] = mean(map(get_rank  ,reward,rewards,len(actions)))
-
-            if self._time: out.update(predict_time=predict_time, learn_time=learn_time)
-            if self._prob and prob is not None: out.update(probability=prob)
-
-            out.update(extra)
-            out.update(learning_info)
+            if learning_info:
+                out.update(learning_info)
+                learning_info.clear()
 
             yield out
-            learning_info.clear()
 
 class SimpleLearnerInfo(LearnerTask):
     """Describe a Learner using its name and hyperparameters."""
@@ -226,7 +234,7 @@ class ClassEnvironmentInfo(EnvironmentTask):
     finished. To make the most of this Task sklearn should be installed.
     """
 
-    def process(self, environment: Environment, interactions: Iterable[SimulatedInteraction]) -> Mapping[Any,Any]:
+    def process(self, environment: Environment, interactions: Iterable[Interaction]) -> Mapping[Any,Any]:
 
         #sources:
         #[1]: https://arxiv.org/pdf/1808.03591.pdf (lorena2019complex)
@@ -236,7 +244,11 @@ class ClassEnvironmentInfo(EnvironmentTask):
 
         #[3] found that information theoretic measures and landmarking measures are most important
 
-        contexts,actions,rewards = zip(*[ (i['context'], i['actions'], i['rewards']) for i in interactions ])
+        interactions = peek_first(interactions)[1]
+        
+        if not interactions: return {}
+
+        contexts, _ ,rewards = zip(*[ (i['context'], i['actions'], i['rewards']) for i in interactions ])
         env_stats = {}
 
         X = [ InteractionsEncoder('x').encode(x=c) for c in contexts ]
@@ -276,7 +288,7 @@ class ClassEnvironmentInfo(EnvironmentTask):
         env_stats["mutual_XY_info_rank1"   ] = mutual_XY_infos[0]
         env_stats["mutual_XY_info_rank2"   ] = mutual_XY_infos[1] if len(mutual_XY_infos) > 1 else None
         env_stats["equivalent_num_X_attr"  ] = entropy_Y/mean(mutual_XY_infos) if mean(mutual_XY_infos) else None #[2,3]
-        env_stats["noise_signal_ratio"     ] = (mean(entropy_X)-mean(mutual_XY_infos))/mean(mutual_XY_infos) if mean(mutual_XY_infos) else None #[2] 
+        env_stats["noise_signal_ratio"     ] = (mean(entropy_X)-mean(mutual_XY_infos))/mean(mutual_XY_infos) if mean(mutual_XY_infos) else None #[2]
 
         env_stats["max_fisher_discrim"    ] = self._max_fisher_discriminant_ratio(X, Y) #[1]
         #env_stats["max_fisher_discrim_dir"] = self._max_directional_fisher_discriminant_ratio(X, Y) #[1] (this dies on large feature)
@@ -387,7 +399,7 @@ class ClassEnvironmentInfo(EnvironmentTask):
 
         IR = (k-1)/k*sum([c/(n-c) for c in counts])
         return 1 - 1/IR
-    
+
     def _max_fisher_discriminant_ratio(self, X, Y) -> float:
         #Equation (3) in [1]
         #needs more testing
