@@ -2,24 +2,23 @@ from collections import abc
 from math import isclose
 from abc import ABC, abstractmethod
 from itertools import repeat
-from typing import Any, Sequence, Union, Tuple, Callable, Mapping, Optional
+from typing import Any, Sequence, Union, Tuple, Callable, Mapping, Optional, Type
 
 from coba.exceptions import CobaException
 from coba.random import CobaRandom
-from coba.environments import Context, Action, Actions
-from coba.primitives import Batch
+from coba.primitives import Context, Action, Actions, AIndex
+from coba.primitives import Batch, Dense, Sparse, HashableDense, HashableSparse
 
 kwargs = Mapping[str,Any]
 Score  = float
-Index  = int
 PMF    = Sequence[float]
-PDF    = Callable[[Union[Action,Index]],float]
+PDF    = Callable[[Union[Action,AIndex]],float]
 
 class Probs(list):
     pass
 
 class ActionScore(tuple):
-    def __new__(self, action: Union[Action,Index], score: Score):
+    def __new__(self, action: Union[Action,AIndex], score: Score):
         return tuple.__new__(ActionScore, (action, score))
 
 Prediction = Union[
@@ -27,7 +26,7 @@ Prediction = Union[
     Probs,
     ActionScore,
     Tuple[PDF                       , kwargs],
-    Tuple[Union[Action,Index],Score , kwargs],
+    Tuple[Union[Action,AIndex],Score, kwargs],
     Tuple[ActionScore               , kwargs],
 ]
 
@@ -64,7 +63,7 @@ class Learner(ABC):
     def learn(self,
         context: Context,
         actions: Actions,
-        action: Union[Action,Index],
+        action: Union[Action,AIndex],
         feedback: Union[float,Any],
         score: float,
         **kwargs) -> None:
@@ -96,6 +95,7 @@ class SafeLearner(Learner):
         self._rng         = CobaRandom(seed)
         self._batched     = None
         self._batched_lrn = None
+        self._batched_mjr = None
         self._pred_type   = None #1==PDF,2==PMF,3==Action/Score
         self._with_info   = None
         self._learn_type  = None #3==current,2==old with info,1==old without info
@@ -125,45 +125,53 @@ class SafeLearner(Learner):
 
         return params
 
-    def predict(self, context: Context, actions: Actions) -> Tuple[Union[Index,Action],Score,kwargs]:
+    def predict(self, context: Context, actions: Actions) -> Tuple[Union[AIndex,Action],Score,kwargs]:
 
         pred      = self._safe_predict(context,actions)
         pred_type = self._pred_type
         batched   = self._batched
 
-        if pred_type is None: # this happens on the first call
-            if batched: old_pred,pred = pred,pred[0]
-            is_discrete = 0 < len(actions) and len(actions) < float('inf')
-            pred_type = self.get_type(pred, is_discrete) or self.get_inferred_type(pred, actions)
-            with_info = self.has_info(pred,pred_type)
-            self._pred_type = pred_type
-            self._with_info = with_info
-            self._info_dict = with_info and isinstance(pred[-1],dict) 
-            if batched: pred = old_pred
+        #if not batched everything just works
+
+        #if batched and what we are given are pmfs
+            #we need the prediction orderd by batch first
+
+        #if batched and what we are given are actions cores
+            #we need the prediction ordered by batch second
+
+        if pred_type is None and not batched: # first call only
+            self._determine_pred_format(pred,actions)
+            pred_type = self._pred_type
+
+        elif pred_type is None and batched:
+            if self._batched_mjr:
+                test_pred = pred[0] 
+            else:
+                test_pred = [p[0] if not isinstance(p,dict) else {k:v[0] for k,v in p.items()} for p in pred]
+            self._determine_pred_format(test_pred,actions[0])
+            pred_type = self._pred_type
 
         if not self._with_info:
             pred_info = {}
-
+        elif self._info_dict:
+            pred_info = {k:[p[-1][k] for p in pred] for k in pred[0][-1]} if self._batched_mjr else pred[-1]
         else:
-            if not batched:
-                pred_info = pred[-1] if self._info_dict else {'_':pred[-1]}
-            if batched:
-                pred_info = [p[-1] if self._info_dict else {'_':p[-1]} for p in pred]
-                pred_info = {k:[ i[k] for i in pred_info] for k in pred_info[0]}
+            pred_info = {'_': [p[-1] for p in pred]} if batched else {'_': pred[-1]}
 
         if self._pred_type == 2:
-            pmf = pred[0] if pred_info else pred
             if not batched:
+                pmf = pred[0] if pred_info else pred
                 action,score = self._get_pmf_action_score(self._rng,pmf,actions)
             else:
+                pmf = [p[0] if pred_info else p for p in pred] if self._batched_mjr else pred[0]
                 action,score = list(zip(*map(self._get_pmf_action_score, repeat(self._rng), pmf, actions)))
         else:
-            if not batched:
-                action,score = pred[:2]
-            else:
-                action,score = zip(*[p[:2] for p in pred])
+            action,score = list(zip(*[p[:2] for p in pred])) if self._batched_mjr else pred[:2]
 
-        return action, score, pred_info
+        if batched:
+            return Batch(action), Batch(score), pred_info
+        else:
+            return action, score, pred_info
 
     def learn(self, context, actions, action, reward, probability, **kwargs) -> None:
 
@@ -186,22 +194,31 @@ class SafeLearner(Learner):
                     all_failed = True
                 if all_failed: raise
 
-    def get_type(self,pred,is_discrete) -> Optional[int]:
+    def _determine_pred_type(self,pred,is_discrete) -> Optional[int]:
         if self._is_type_1(pred): raise CobaException("PDF predictions are currently not supported.")
         if self._is_type_2(pred,is_discrete): return 2
         if self._is_type_3(pred,is_discrete): return 3
         return None
 
-    def has_info(self,pred,pred_type) -> Mapping[Any,Any]:
+    def _determine_has_info(self,pred,pred_type) -> Mapping[Any,Any]:
         if pred_type == 2:
             return not isinstance(pred[0],(int,float))
         else: #pred_type==3
             return len(pred) == 3
 
+    def _determine_pred_format(self, pred, actions):
+
+        is_discrete = 0 < len(actions) and len(actions) < float('inf')
+        pred_type = self._determine_pred_type(pred, is_discrete) or self.get_inferred_type(pred, actions)
+        with_info = self._determine_has_info(pred,pred_type)
+
+        self._pred_type = pred_type
+        self._with_info = with_info
+        self._info_dict = with_info and isinstance(pred[-1],dict) 
+
     def _safe_predict(self, context, actions):
 
         batched = self._batched
-        batched_lrn = self._batched_lrn
 
         if batched is None: 
             batched = isinstance(context,Batch)
@@ -209,25 +226,39 @@ class SafeLearner(Learner):
 
         if not batched:
             return self._learner.predict(context,actions)
+
         if batched:
+            batched_lrn = self._batched_lrn
             if batched_lrn == True:
                 return self._learner.predict(context,actions)
             elif batched_lrn == False:
-                return [self._learner.predict(c, a) for c,a in zip(context,actions)]
-            else:
+                return list(map(self._learner.predict,context,actions))
+            elif batched_lrn is None:
                 try:
+                    batch_size = len(context)
                     pred = self._learner.predict(context,actions)
-                    #I'm not sure if this is air tight but I think it is?
-                    pred_is_appropriately_batched = len(context) == len(pred) and not isinstance(pred[0],(int,float))
-                    if pred_is_appropriately_batched:
-                        self._batched_lrn = True
-                        return pred
+                    n_rows = len(pred)
+                    n_cols = len(pred[0])
+
+                    self._batched_lrn = True
+
+                    if n_rows != n_cols:
+                        self._batched_mjr = len(pred) == batch_size
                     else:
-                        self._batched_lrn = False
-                        return [self._learner.predict(c, a) for c,a in zip(context,actions)]
-                except:
+                        #The major order of pred is not determinable. So we
+                        #now do a small "test" to determine the major order.
+                        test_pred = self._learner.predict(Batch([context[0]]),Batch([actions[0]]))
+                        self._batched_mjr = len(test_pred) == 1
+                    return pred
+
+                except Exception as e:
                     self._batched_lrn = False
-                    return [self._learner.predict(c, a) for c,a in zip(context,actions)]
+                    self._batched_mjr = True
+                    try:
+                        return list(map(self._learner.predict,context,actions))
+                    except:
+                        pass
+                    raise
 
     def _safe_learn(self,context,actions,action,reward,probability,kwargs):
         if self._learn_type==3:
@@ -315,3 +346,24 @@ class SafeLearner(Learner):
 
     def __str__(self) -> str:
         return self.full_name
+
+def requires_hashables(cls:Type[Learner]):
+
+    def make_hashable(item):
+        if isinstance(item,Dense): return HashableDense(item)
+        if isinstance(item,Sparse): return HashableSparse(item)
+        return item
+
+    old_predict = cls.predict
+    old_learn   = cls.learn
+
+    def new_predict(self,c,A):
+        return old_predict(self,make_hashable(c),list(map(make_hashable,A)))
+
+    def new_learn(self,c,A,a,r,p,**kwargs):
+        old_learn(self,make_hashable(c),list(map(make_hashable,A)), make_hashable(a),r,p,**kwargs)
+
+    cls.predict = new_predict
+    cls.learn   = new_learn
+
+    return cls
